########################################################################################################################

__doc__ = \
    """
This is a variant of Victor for MPro that uses data from PostEra
    """

########################################################################################################################


from typing import List, Optional
from warnings import warn

import pandas as pd
import pyrosetta
from rdkit import Chem

from ..demo import MPro as data  # noqa this is basically a package, but actually a class
from ..victor import Victor
from .dataframe import fetch_postera, read_postera


class MProVictor(Victor):
    constraint_function_type = 'FLAT_HARMONIC'

    @classmethod
    def from_hit_codes(cls, hit_codes: List[str], **options):
        hits = [data.get_mol(xnumber) for xnumber in hit_codes]
        return cls(hits=hits, **options)

    @staticmethod
    def pose_fx(pose: pyrosetta.Pose):
        """
        Histidine protonated on epsilon — HIE.
        """
        pdb2pose = pose.pdb_info().pdb2pose
        r = pdb2pose(res=41, chain='A')
        MutateResidue = pyrosetta.rosetta.protocols.simple_moves.MutateResidue
        MutateResidue(target=r, new_res='HIS').apply(pose)

    @staticmethod
    def poised_pose_fx(pose: pyrosetta.Pose):
        """
        Histidine protonated on delta (HID) and cysteine in thiolate.
        """
        # Americans spell poise as poise not poize.
        pdb2pose = pose.pdb_info().pdb2pose
        r = pdb2pose(res=41, chain='A')
        MutateResidue = pyrosetta.rosetta.protocols.simple_moves.MutateResidue
        MutateResidue(target=r, new_res='HIS_D').apply(pose)
        r = pdb2pose(res=145, chain='A')
        MutateResidue(target=r, new_res='CYZ').apply(pose)

    def __init__(self, category:Optional[str]=None, **options):
        # this category flag is solely for Mpro?
        # it stems from the Moonshot file.
        self.category = category
        if category == 'noncolavent':
            fx = self.__class__.poised_pose_fx
        else:
            fx = self.__class__.pose_fx
        # --------------------
        defaults = dict(
            pdb_block=data.get_template(),
            ligand_resn='LIG',
            ligand_resi='1B',
            covalent_resn='CYS',
            covalent_resi='145A',
            extra_protein_constraint='AtomPair  SG  145A  NE2  41A HARMONIC 3.5 0.2\n',
            pose_fx=fx  # from the above.
        )
        super().__init__(**{**defaults, **options})

    def _determine_extras(self, smiles):
        defaults = dict()
        if self.category not in (None, 'noncovalent') and '_' in self.category:
            cname, rxd = self.category.split('_')
            if rxd == 'noncovalent':
                wd = [wd for wd in self.warhead_definitions if wd['name'] == cname][0]
                mol = Chem.MolFromSmiles(smiles)
                nc = Chem.MolFromSmiles(wd['noncovalent'])
                atomnames = dict(zip(mol.GetSubstructMatch(nc), wd['noncovalent_atomnames']))
                extra_constraint = 'AtomPair  SG  145A  CX   1B HARMONIC 3.2 0.5\n'
                extra_constraint += wd['constraint']
                defaults = dict(atomnames=atomnames,
                                extra_ligand_constraint=extra_constraint)
        return defaults

    #self.combine(**options) unchanged.

    def place(self, *args, **options):
        defaults = self._determine_extras(options['smiles'] if 'smiles' in options else args[0])
        # singledispatchmethod does not like named args
        kwargs = {**defaults, **options}
        return super().place(*args, **kwargs)

    # ======= postera csv file ops =====================================================================================

    @classmethod
    def fetch_postera(cls):
        __doc__ = fetch_postera.__doc__
        warn('method moved to mpro_data (as function)', category=DeprecationWarning)
        return fetch_postera()

    @classmethod
    def analyse_postera(cls):
        raise NotImplementedError('Incomplete functionality')

    @classmethod
    def from_postera_row(cls, row: pd.Series, results:Optional=None):
        if row.fragments == 'x0072' or str(row.fragments) == 'nan':
            # these are not hit inspired.
            cls.journal.error(f'No valid parent hits for {row.CID}.')
            return None
        elif results and row.CID in results:
            cls.journal.info(f'{row.CID} has already been done.')
            return None
        elif row.covalent_warhead in (False, 'False', 'false'):
            # parse
            return cls.from_hit_codes(hit_codes=row.fragments.split(','),
                                      category='noncolavent')\
                      .place(long_name=row.CID,
                             smiles=row.SMILES)
        elif row.category not in ('Acrylamide', 'Chloroacetamide', 'Vinylsulfonamide', 'Nitrile'):
            cls.journal.info(f'What is {row["CID"]}? Treating like a non-covalent.')
            return cls.from_hit_codes(hit_codes=row.fragments.split(','),
                                      category='noncolavent')\
                      .place(long_name=row.CID,
                             smiles=row.SMILES)
        else:
            return cls.from_hit_codes(hit_codes=row.fragments.split(','),
                                      category=row.category)\
                      .place(long_name=row.CID,
                             smiles=row.SMILES)

#### Update the dataset with constraints specific for MPro

for cname, con in [('chloroacetamide', 'AtomPair  H  145A  OY  1B HARMONIC 2.1 0.2\n'),
                   ('nitrile', 'AtomPair  H  145A  NX  1B HARMONIC 2.1 0.2\n'),
                   ('acrylamide', 'AtomPair  H  143A  OZ  1B HARMONIC 2.1 0.2\n'),
                   ('vinylsulfonamide', 'AtomPair  H  143A  OZ1 1B HARMONIC 2.1 0.2\n'),
                    ('bromoalkyne', 'AtomPair  H  145A  CY  1B HARMONIC 2.1 0.2\n'),
                   ]:
    MProVictor.add_constraint_to_warhead(name=cname, constraint=con)









import copy
import random
from typing import Optional, Dict, Union

from rdkit import Chem

from fragmenstein import Victor

class MultiVictorPlacement():
    """
    MultiVictorPlacement is a class to execute multiple times victor.place over the same smile using different seeds.
    The idea is to be able to do something more like docking for those cases in which the parent hits do not
    explain an important part of the molecule
    """

    def __init__(self, random_seed=None, **victor_init_args):
        self.random_seed = random_seed
        self.victor_init_args = victor_init_args

        self.placed_victors = []

    def place(self,  smiles: str, number_runs: int= 10,
              long_name: str = 'ligand',
              merging_mode='expansion',
              atomnames: Optional[Dict[int, str]] = None,
              custom_map: Optional[Dict[str, Dict[int, int]]] = None,
              extra_ligand_constraint: Union[str] = None):
        """
        Places a followup (smiles) into the protein based upon the hits. Obtains number_runs solutions.
        :param smiles: smiles of followup, optionally covalent (_e.g._ ``*CC(=O)CCC``)
        :param number_runs: the number of Victor objects to create and place the smiles
        :param long_name: gets used for filenames so will get corrected
        :param merging_mode:
        :param atomnames: an optional dictionary that gets used by ``Params.from_smiles``
        :param custom_map: see Monster.place
        :param extra_ligand_constraint:
        :return:
        """
        assert number_runs >= 1, "Error, at least one placement required"
        current_randomState = None
        if self.random_seed:
            current_randomState = random.getstate()
            random.seed(self.random_seed)
        rseeds = [random.randint(0, int(1e6)) for i in range(number_runs)]
        if current_randomState is not None:
            random.setstate(current_randomState)

        i = 0
        victor = Victor(monster_random_seed=rseeds[i], **self.victor_init_args)
        victor.place(smiles, long_name=long_name+str(i), merging_mode= merging_mode,
                     custom_map=custom_map, atomnames=atomnames,
                     extra_ligand_constraint=extra_ligand_constraint)
        victor.runNumber = i
        ddG = victor.ddG
        monster = victor.monster
        self.placed_victors = [(ddG, victor)]

        for i in range(1, number_runs):
            monster.sample_new_conformation(rseeds[i])
            victor = Victor(monster_random_seed=rseeds[i], **self.victor_init_args)
            victor._prepare_args_for_placement(smiles=smiles, long_name = long_name+str(i), merging_mode= merging_mode,
                                               custom_map=custom_map,
                                               atomnames=atomnames, extra_ligand_constraint=extra_ligand_constraint)
            victor.monster = monster
            victor._calculate_placement_thermo()
            self.placed_victors += [(victor.ddG, victor)]
        self.placed_victors = sorted(self.placed_victors)

    def retrieve_best_victor(self):
        return self.placed_victors[0][1]

    def retrieve_scores(self):
        return list(zip(*self.placed_victors))[0]



default_settings_yaml = """
# These are the default settings for fragmenstein.
# To overide please define $FRAGMENSTEIN_SETTINGS as a yaml file.
# or pass an environment variable _prior_ to import 
# e.g. ff_constraint becomes $FRAGMENSTEIN_FF_CONSTRAINT.
# These will have priority over the defaults.
# Note there are no safeguards against typos.
# For the command line interface, see fragmenstein/_cli_defaults.py

# General settings
work_path: output
monster_average_position: false
monster_throw_on_discard: false
ff_minisation: true

# During the RDKit minisation, how much lee-way to give an atom before it gets penalised.
ff_max_displacement: 0.1

# During the RDKit minisation, how much to penalise an atom that is too far from its ideal position.
ff_constraint: 5.

# During the RDKit minisation, how many iterations to run.
ff_max_iterations: 200

# During the RDKit minisation, use the neighbourhood to constrain the molecule.
ff_use_neighborhood: true
ff_neighborhood: 6.0
ff_allow_lax: true
ff_prevent_cis: true

# For Wictor, weird things happen if True
ff_minimise_ideal: false

# OpenMM settings
mm_restraint_k: 1000.0
mm_tolerance: 10.0  # mmu.kilocalorie_per_mole / (mmu.nano * mmu.meter)
mm_max_iterations: 0 # 0 is infinite
mm_mobile_radius: 8.0  # mmu.angstrom
"""

import os, yaml
from pathlib import Path
from warnings import warn
from ._cli_defaults import cli_default_settings

if not os.environ.get('FRAGMENSTEIN_SETTINGS', None):
    default_settings = yaml.load(default_settings_yaml, Loader=yaml.FullLoader)
elif Path(os.environ['FRAGMENSTEIN_SETTINGS']).exists():
        with open(os.environ['FRAGMENSTEIN_SETTINGS']) as r:
            default_settings = yaml.load(r, Loader=yaml.FullLoader)
else:
    raise ValueError(f'FRAGMENSTEIN_SETTINGS={os.environ["FRAGMENSTEIN_SETTINGS"]} does not exist as a file')

for variable in os.environ:
    if 'FRAGMENSTEIN_' in variable:
        key = variable.replace('FRAGMENSTEIN_', '').lower()
        if key not in default_settings and key not in cli_default_settings:
            # there should be some kind of check here if a CLI variable is called within Python.
            warn(f'Environment variable {variable} is not a valid setting and will likely be ignored')
        key_type = type(default_settings.get(key, ''))
        default_settings[key] = key_type(os.environ[variable])



"""
The ``mcs_mapping`` module contains the classes and functions for mapping with the MCS algorithm,
restricted to a mapping that must be obeyed.
"""


from .types import IndexMap, BasicFMCSMode, ExtendedFMCSMode
from .compare_atom import VanillaCompareAtoms, SpecialCompareAtoms
from .utils import flip_mapping, transmute_FindMCS_parameters


########################################################################################################################

__doc__ = \
    """
base methods
    """

########################################################################################################################
import os.path

from .pyrosetta_import import pyrosetta  # the real mcCoy or a mock.
from ._igor_base import _IgorBase

from typing import Dict, List, Optional, Tuple, Union, Sequence

from warnings import warn


class _IgorInit(_IgorBase):
    atom_pair_constraint = 10
    angle_constraint = 10
    coordinate_constraint = 1
    fa_intra_rep = 0.005

    # ============= Init ===============================================================================================

    def __init__(self,
                 pose: pyrosetta.Pose,
                 constraint_file: str,
                 ligand_residue: Union[str, int, Tuple[int, str], pyrosetta.Vector1] = 'LIG',
                 key_residues: Union[None, Sequence[Union[int, str, Tuple[int, str]]], pyrosetta.Vector1] = None):
        """
        Given a pose with a blended ligand at ligand residue. Load it (ready for minimisation).

        :param pose: pose.
        :param constraint_file: filename
        :param ligand_residue: ligands -see class docstring
        :param key_residues: multiple entries -see class docstring
        """
        self.pose = pose  #: pyrosetta.Pose
        # virtualroot
        if pose.residue(self.pose.total_residue()).name3() != 'XXX':
            pyrosetta.rosetta.core.pose.addVirtualResAsRoot(self.pose)
        pyrosetta.create_score_function('ref2015')(self.pose)
        self.constraint_file = constraint_file  #: str
        self.ligand_residue: List[int] = self._parse_residue(ligand_residue)
        self.key_residues: List[int] = self._parse_key_residues(key_residues)

    @classmethod
    def from_pdbblock(cls,
                      pdbblock: str,
                      params_file: str,
                      constraint_file: str,
                      ligand_residue: Union[str, int, Tuple[int, str], pyrosetta.Vector1] = 'LIG',
                      key_residues: Union[None, Sequence[Union[int, str, Tuple[int, str]]], pyrosetta.Vector1] = None):


        pose = pyrosetta.Pose()
        params_paths = pyrosetta.rosetta.utility.vector1_string()
        params_paths.extend([params_file])
        pyrosetta.generate_nonstandard_residue_set(pose, params_paths)
        pyrosetta.rosetta.core.import_pose.pose_from_pdbstring(pose, pdbblock)
        return cls(pose, constraint_file, ligand_residue, key_residues)

    @classmethod
    def from_pdbfile(cls,
                     pdbfile: str,
                     params_file: str,
                     constraint_file: str,
                     ligand_residue: Union[str, int, Tuple[int, str], pyrosetta.Vector1] = 'LIG',
                     key_residues: Union[None, Sequence[Union[int, str, Tuple[int, str]]], pyrosetta.Vector1] = None):
        """
        Given a PDB with the ligand load it.

        :param pdbfile: pdb file
        :param params_file: params file
        :param constraint_file: filename
        :param ligand_residue: ligand -see class docstring
        :param key_residues: multiple entries -see class docstring
        :return:
        """
        for filename in (pdbfile, constraint_file, params_file):
            if not os.path.exists(filename):
                raise FileNotFoundError(f'{filename} does not exist')
            if os.stat(filename).st_size < 10:
                raise ValueError(f'{filename} is only {os.stat(filename).st_size} bytes long')
        pose = pyrosetta.Pose()
        params_paths = pyrosetta.rosetta.utility.vector1_string()
        params_paths.extend([params_file])
        pyrosetta.generate_nonstandard_residue_set(pose, params_paths)
        pyrosetta.rosetta.core.import_pose.pose_from_file(pose, pdbfile)
        return cls(pose, constraint_file, ligand_residue, key_residues)

    # ============= Private methods for init ===========================================================================

    def _parse_residue(self, residue: Union[int, str, Tuple[int, str], pyrosetta.Vector1]) -> List[int]:
        """
        Parse a residue into a list of pose residue indices.
        This will be called to fill ``self.ligand_residue`` and ``self.key_residues``.

        :param residue: residue to parse
        :type residue: Union[int, str, Tuple[int, str], pyrosetta.Vector1]
        :return: list of pose residue indices
        """
        parsed: List[int] = []
        if residue is None:
            ## assuming it is LIG then.
            ligand_selector = pyrosetta.rosetta.core.select.residue_selector.ResidueNameSelector()
            ligand_selector.set_residue_name3('LIG')
            m = self._vector2residues(ligand_selector.apply(self.pose))  # noqa its in Igor_min
            if len(m) > 1:
                warn('There are many residues called LIG!. Please consider specifying the name3/resn/resi yourself!')
            for r in m:
                parsed.append(r)
        elif isinstance(residue, int):
            # it is a pose residue index.
            parsed.append(residue)
        elif isinstance(residue, str) and residue.isdigit():
            parsed.append(int(residue))
        elif isinstance(residue, str) and residue[:-1].isdigit():
            # it is a pdb residue index + chain e.g. 23A.
            res = int(residue[:-1])
            chain = residue[-1]
            r = self.pose.pdb_info().pdb2pose(res=res, chain=chain)
            parsed.append(r)
        elif isinstance(residue, str) and len(residue) in (3, 4) and residue.upper() == residue:
            # it is a residue name, such as LIG
            ligand_selector = pyrosetta.rosetta.core.select.residue_selector.ResidueNameSelector()
            ligand_selector.set_residue_name3(residue)
            for r in [i + 1 for i, v in enumerate(ligand_selector.apply(self.pose)) if v == 1]:
                parsed.append(r)
        elif isinstance(residue, tuple) and len(residue) == 2:
            parsed = self.pose.pdb_info().pdb2pose(res=residue[0], chain=residue[1])
        # elif isinstance(residue, pyrosetta.Vector1): will raise an error
        # as the latter is a function. that redirects to `vector1_int`, `vector1_double` etc.
        elif hasattr(residue, '__class__') and 'vector1' in residue.__class__.__name__:
            for r in [int(i + 1) for i, v in enumerate(residue) if int(v) == 1]:
                parsed.append(r)
        else:
            raise ValueError(f'No idea what {residue} is.')
        parsed = [p for p in parsed if p != 0]
        assert len(parsed), (f'There is no {residue} in pose. '+
                            'This can happen if user provided PDB residues not exist or '+
                             'are invalid (Calpha only, no density etc). Suggestion: check `victor.covalent_resi`.')
        return parsed

    def _parse_key_residues(self,
                            key_residues: Union[None, Sequence[Union[int, str, Tuple[int, str]]],
                                                pyrosetta.Vector1]):
        """
        Parse a list of residues into a list of pose residue indices that will be filling ``self.key_residues``.
        Does not include the ligand residue(s).
        """
        parsed: List[int] = []
        residue = self.pose.residue(self.ligand_residue[0])
        if key_residues is None and residue.n_current_residue_connections():
            parsed = [residue.connect_map(1).resid()]
        elif key_residues is None:
            parsed = [1]
        else:
            for k in key_residues:
                parsed.extend(self._parse_residue(k))
        return parsed




# https://stackoverflow.com/questions/19425736/how-to-redirect-stdout-and-stderr-to-logger-in-python

class LoggerWriter:
    """
    Write stderr to logger.
    See capture_rdkit_log method of Victor
    where this happens:
    sys.stderr = LoggerWriter(cls.journal.warning)
    """
    def __init__(self, writer):
        self._writer = writer
        self._msg = ''

    def write(self, message):
        self._msg += str(message)
        # this causes an infinite recursion and
        # I am not sure why it was added
        # while '\n' in self._msg:
        #     pos = self._msg.find('\n')
        #     self._writer(self._msg[:pos])
        #     self._msg = self._msg[pos + 1:]

    def flush(self):
        if self._msg != '':
            self._writer(self._msg)
            self._msg = ''
